/*
 * Copyright 2013 Canonical Ltd.
 *
 * This file is part of powerd.
 *
 * powerd is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3.
 *
 * powerd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <glib.h>
#include <glib-object.h>
#include <gio/gio.h>

#include "powerd-cli.h"

static int set_user_autobrightness(gboolean enable)
{
    GDBusProxy *proxy;
    GVariant *ret = NULL;
    GError *error = NULL;

    proxy = cli_get_dbus_proxy();
    if (!proxy) {
        fprintf(stderr, "Error: Could not get dbus proxy\n");
        return -EIO;
    }

    ret = g_dbus_proxy_call_sync(proxy, "userAutobrightnessEnable",
                                 g_variant_new("(b)", enable),
                                 G_DBUS_CALL_FLAGS_NONE, -1, NULL, &error);
    if (!ret) {
        fprintf(stderr, "Error: Failed to set autobrightness: %s\n",
                error->message);
        g_error_free(error);
        return -EIO;
    }

    g_variant_unref(ret);
    return 0;
}

static int set_user_brightness(int brightness)
{
    GDBusProxy *proxy;
    GVariant *ret = NULL;
    GError *error = NULL;

    proxy = cli_get_dbus_proxy();
    if (!proxy) {
        fprintf(stderr, "Error: Could not get dbus proxy\n");
        return -EIO;
    }
    
    ret = g_dbus_proxy_call_sync(proxy, "setUserBrightness",
                                 g_variant_new("(i)", brightness),
                                 G_DBUS_CALL_FLAGS_NONE, -1, NULL, &error);
    if (!ret) {
        fprintf(stderr, "Error: Failed to set user brightness: %s", error->message);
        g_error_free(error);
        return -EIO;
    }

    g_variant_unref(ret);
    return 0;
}

static int get_brightness_params(int *min, int *max, int *dflt,
                                 gboolean *ab_supported)
{
    GDBusProxy *proxy;
    GVariant *ret = NULL;
    GError *error = NULL;

    proxy = cli_get_dbus_proxy();
    if (!proxy) {
        fprintf(stderr, "Error: Could not get dbus proxy\n");
        return -EIO;
    }

    ret = g_dbus_proxy_call_sync(proxy, "getBrightnessParams", NULL,
                                 G_DBUS_CALL_FLAGS_NONE, -1, NULL, &error);
    if (!ret) {
        fprintf(stderr, "Error: Failed to get brightness paramaters: %s",
                error->message);
        g_error_free(error);
        return -EIO;
    }

    g_variant_get(ret, "((iiib))", min, max, dflt, ab_supported);
    g_variant_unref(ret);

    return 0;
}

static int run_cmd_autobrightness(int argc, char *argv[])
{
    gboolean enable;
    const char *arg;

    if (argc != 2) {
        /* XXX: Print usage */
        return -EINVAL;
    }
    arg = argv[1];

    if (!strcmp(arg, "enable")) {
        enable = TRUE;
    } else if (!strcmp(arg, "disable")) {
        enable = FALSE;
    } else {
        fprintf(stderr, "Error: Invalid autobrightness state\n");
        /* XXX: Print usage */
        return -EINVAL;
    }

    return set_user_autobrightness(enable);
}

const struct cli_cmd autobrightness_cmd = {
    .name = "autobrightness",
    .help_text = "Enable or disable automatic brightness control.",
    .run = run_cmd_autobrightness,
};

static int run_cmd_brightness(int argc, char *argv[])
{
        long brightness;
        char *arg, *endp;

        if (argc != 2) {
            /* XXX: Usage */
            return -EINVAL;
        }
        arg = argv[1];

        errno = 0;
        brightness = strtol(arg, &endp, 10);
        if (errno == ERANGE || *endp != '\0' ||
            brightness < 0 || brightness > INT_MAX) {
            fprintf(stderr, "Error: Invalid brightness %s\n", arg);
            /* XXX: Usage */
            return -EINVAL;
        }

        return set_user_brightness((int)brightness);
}

const struct cli_cmd brightness_cmd = {
    .name = "brightness",
    .help_text = "Set screen brightness.",
    .run = run_cmd_brightness,
};

static int run_cmd_brightness_params(int argc, char *argv[])
{
    int min, max, dflt, ab;
    int ret;

    if (argc != 1) {
        /* XXX: Usage */
        return -EINVAL;
    }
        
    ret = get_brightness_params(&min, &max, &dflt, &ab);
    if (!ret) {
        printf("Minimum Brightness: %d\n"
               "Maximum Brightness: %d\n"
               "Default Brightness: %d\n"
               "Autobrightness:     %ssupported\n",
               min, max, dflt, ab ? "" : "not ");
    }

    return ret;
}

const struct cli_cmd brightness_params_cmd = {
    .name = "brightness-params",
    .help_text = "Get brightness parameters (min, max, etc.).",
    .run = run_cmd_brightness_params,
};
