/*
 * Copyright 2013 Canonical Ltd.
 *
 * This file is part of powerd.
 *
 * powerd is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3.
 *
 * powerd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <errno.h>
#include <glib.h>
#include <glib-object.h>
#include <gio/gio.h>

#include "powerd-cli.h"

struct SysRequestStats {
    const char *owner;
    const char *name;
    unsigned active_count;
    guint64 active_time;
    guint64 max_active_time;
    guint64 active_since;
};

struct DispRequestStats {
    const char *owner;
    const char *name;
    unsigned active_count;
    guint64 active_time;
    guint64 max_active_time;
    guint64 active_since;
    guint64 disp_on_time;
    guint64 disp_on_since;
    guint64 flag_on_time[POWERD_NUM_DISPLAY_FLAGS];
    guint64 flag_on_since[POWERD_NUM_DISPLAY_FLAGS];
};

static GArray *get_sys_request_stats(GDBusProxy *proxy)
{
    GVariant *ret, *item;
    GVariantIter *iter;
    GArray *retarray;
    GError *error;

    retarray =  g_array_new(FALSE, FALSE, sizeof(struct SysRequestStats));

    error = NULL;
    ret = g_dbus_proxy_call_sync(proxy, "getSysRequestStats", NULL,
                                 G_DBUS_CALL_FLAGS_NONE, -1, NULL, &error);
    if (ret == NULL) {
        cli_warn("getSysRequestStats failed: %s", error->message);
        g_error_free(error);
    } else {
        g_variant_get(ret, "(a(ssuttt))", &iter);
        while ((item = g_variant_iter_next_value (iter))) {
            struct SysRequestStats stats;
            g_variant_get_child(item, 0, "s", &stats.owner);
            g_variant_get_child(item, 1, "s", &stats.name);
            g_variant_get_child(item, 2, "u", &stats.active_count);
            g_variant_get_child(item, 3, "t", &stats.active_time);
            g_variant_get_child(item, 4, "t", &stats.max_active_time);
            g_variant_get_child(item, 5, "t", &stats.active_since);
            g_array_append_val(retarray, stats);
            g_variant_unref(item);
        }
        g_variant_unref(ret);
    }

    return retarray;
}

static void print_sys_request_stats(GArray *stats)
{
    int i;
    struct SysRequestStats *stat;

    printf("System Request Statistics:\n");

    if (stats->len == 0) {
        printf("  None\n");
    } else {
        printf("  %-16.16s %-20.20s %-8.8s %-16.16s %-16.16s %-16.16s\n",
               "", "", "Active", "", "Max Active", "");
        printf("  %-16.16s %-20.20s %-8.8s %-16.16s %-16.16s %-16.16s\n",
               "Owner", "Name", "Count", "Active Time", "Time", "Active Since");
        for (i = 0; i < stats->len; i++) {
            stat = &g_array_index(stats, struct SysRequestStats, i);
            printf("  %-16.16s %-20.20s %-8u %-16.6f %-16.6f %-16.6f\n",
                   stat->owner, stat->name, stat->active_count,
                   (double)stat->active_time / 1000000.0f,
                   (double)stat->max_active_time / 1000000.0f,
                   (double)stat->active_since / 1000000.0f);
        }
    }
}

static GArray *get_disp_request_stats(GDBusProxy *proxy)
{
    GVariant *ret, *item;
    GVariantIter *iter;
    GArray *retarray;
    GError *error;

    retarray =  g_array_new(FALSE, FALSE, sizeof(struct DispRequestStats));

    error = NULL;
    ret = g_dbus_proxy_call_sync(proxy, "getDispRequestStats", NULL,
                                 G_DBUS_CALL_FLAGS_NONE, -1, NULL, &error);
    if (ret == NULL) {
        cli_warn("getDispRequestStats failed: %s", error->message);
        g_error_free(error);
    } else {
        g_variant_get(ret, "(a(ssutttttatat))", &iter);
        while ((item = g_variant_iter_next_value (iter))) {
            struct DispRequestStats stats;
            GVariantIter *array_iter;
            guint64 val;
            int i;

            g_variant_get_child(item, 0, "s", &stats.owner);
            g_variant_get_child(item, 1, "s", &stats.name);
            g_variant_get_child(item, 2, "u", &stats.active_count);
            g_variant_get_child(item, 3, "t", &stats.active_time);
            g_variant_get_child(item, 4, "t", &stats.max_active_time);
            g_variant_get_child(item, 5, "t", &stats.active_since);
            g_variant_get_child(item, 6, "t", &stats.disp_on_time);
            g_variant_get_child(item, 7, "t", &stats.disp_on_since);

            g_variant_get_child(item, 8, "at", &array_iter);
            i = 0;
            while (g_variant_iter_loop(array_iter, "t", &val)) {
                if (i >= POWERD_NUM_DISPLAY_FLAGS)
                    break;
                stats.flag_on_time[i++] = val;
            }
            g_variant_iter_free(array_iter);

            g_variant_get_child(item, 9, "at", &array_iter);
            i = 0;
            while (g_variant_iter_loop(array_iter, "t", &val)) {
                if (i >= POWERD_NUM_DISPLAY_FLAGS)
                    break;
                stats.flag_on_since[i++] = val;
            }
            g_variant_iter_free(array_iter);

            g_array_append_val(retarray, stats);
            g_variant_unref(item);
        }
        g_variant_unref(ret);
    }

    return retarray;
}

static void print_disp_request_stats(GArray *stats)
{
    int i, j;
    struct DispRequestStats *stat;

    printf("Display Request Statistics:\n");

    if (stats->len == 0) {
        printf("  None\n");
    } else {
        printf("  %-16.16s %-20.20s %-8.8s %-16.16s %-16.16s %-16.16s %-16.16s %-16.16s",
               "", "", "Active", "", "Max Active", "", "Display On",
               "Display On");
        for (i = 0; i < POWERD_NUM_DISPLAY_FLAGS; i++)
            printf(" Flag %-2d          Flag %-2d         ", i, i);
        printf("\n  %-16.16s %-20.20s %-8.8s %-16.16s %-16.16s %-16.16s %-16.16s %-16.16s",
               "Owner", "Name", "Count", "Active Time", "Time", "Active Since",
               "Time", "Since");
        for (i = 0; i < POWERD_NUM_DISPLAY_FLAGS; i++)
            printf(" On Time          On Since        ");
        printf("\n");
        for (i = 0; i < stats->len; i++) {
            stat = &g_array_index(stats, struct DispRequestStats, i);
            printf("  %-16.16s %-20.20s %-8u %-16.6f %-16.6f %-16.6f %-16.6f %-16.6f",
                   stat->owner, stat->name, stat->active_count,
                   (double)stat->active_time / 1000000.0f,
                   (double)stat->max_active_time / 1000000.0f,
                   (double)stat->active_since / 1000000.0f,
                   (double)stat->disp_on_time / 1000000.0f,
                   (double)stat->disp_on_since / 1000000.0f);
            for (j = 0; j < POWERD_NUM_DISPLAY_FLAGS; j++) {
                printf(" %-16.6f", (double)stat->flag_on_time[j] / 1000000.0f);
                printf(" %-16.6f", (double)stat->flag_on_since[j] / 1000000.0f);
            }
            printf("\n");
        }
    }
}

static int run_stats(int argc, char *argv[])
{
    GDBusProxy *proxy;
    GArray *requests;

    if (argc != 1) {
        /* XXX: Usage */
        return -EINVAL;
    }

    proxy = cli_get_dbus_proxy();
    if (!proxy) {
        fprintf(stderr, "Error: Could not get dbus proxy\n");
        return -EIO;
    }

    requests = get_sys_request_stats(proxy);
    print_sys_request_stats(requests);
    g_array_free(requests, TRUE);

    requests = get_disp_request_stats(proxy);
    print_disp_request_stats(requests);
    g_array_free(requests, TRUE);

    return 0;
}

const struct cli_cmd stats_cmd = {
    .name = "stats",
    .help_text = "Print request statistics.",
    .run = run_stats,
};
