/*
 * Copyright 2013 Canonical Ltd.
 *
 * This file is part of powerd.
 *
 * powerd is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3.
 *
 * powerd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <glib.h>
#include <glib-object.h>
#include <gio/gio.h>

#include "powerd-cli.h"

#define TEST_NUM_SYS_REQUESTS 5
#define TEST_NUM_DISP_REQUESTS (POWERD_NUM_DISPLAY_STATES * 2)

static GMainLoop *main_loop = NULL;
static powerd_cookie_t main_cookie;
static const char *powerd_cli_bus_name;
static char test_dbusname[128];

/* Set to TRUE during tests to silence expected errors */
gboolean silent_errors = FALSE;

static void silence_errors(gboolean silent)
{
    silent_errors = silent;
}

/*
 * Check that dbus owner appears count times in a list of sys/disp
 * requests.
 */
static gboolean check_for_dbus_name(const char *dbusname, int count,
                               GArray *requests, gboolean isSys)
{
    int i;
    int found = 0;
    struct PublicSysRequest *psr;
    struct PublicDispRequest *pdr;

    if (requests == NULL) {
        cli_warn("NULL requests passed to %s", __func__);
        return FALSE;
    }

    for (i = 0; i < requests->len; i++) {
        if (isSys) {
            psr = &g_array_index(requests, struct PublicSysRequest, i);
            if (!strcmp(psr->owner, dbusname))
                found++;
        } else {
            pdr = &g_array_index(requests, struct PublicDispRequest, i);
            if (!strcmp(pdr->owner, dbusname))
                found++;
        }
    }

    if (found != count) {
        cli_debug("Expected %d requests from DbusName (%s), found %d",
            count, dbusname, found);
        return FALSE;
    }

    return TRUE;
}

static void cb_child_watch(GPid pid, gint status, gpointer *data)
{
    GArray *requests;

    /* Make sure that clean-up worked */
    if (test_dbusname[0] != 0) {
        requests = listSysRequests();
        do_test(check_for_dbus_name(test_dbusname, 0, requests, TRUE));
        g_array_free(requests, TRUE);

        requests = listDisplayRequests();
        do_test(check_for_dbus_name(test_dbusname, 0, requests, TRUE));
        g_array_free(requests, TRUE);
    } else {
        printf("Did not find child's dbus name\n");
        printf("    result: FAILED\n");
    }

    /*
     * Make sure clean-up didn't remove our request, note we don't pass
     * the child PID here.
     */
    requests = listSysRequests();
    do_test(check_for_dbus_name(powerd_cli_bus_name, 2, requests, TRUE));
    g_array_free(requests, TRUE);

    /* Close pid */
    g_spawn_close_pid(pid);

    g_main_loop_quit(main_loop);

    /* cleanup */
    do_test(clearSysState(main_cookie) == TRUE);

    requests = listSysRequests();
    do_test(check_for_dbus_name(powerd_cli_bus_name, 0, requests, TRUE));
    g_array_free(requests, TRUE);
}

static gboolean cb_out_watch(GIOChannel *channel, GIOCondition cond,
                             gpointer *data)
{
    gchar *string = NULL;
    gsize size;
    GIOStatus status;

    if (cond == G_IO_HUP) {
        g_io_channel_unref(channel);
        return FALSE;
    }

    status = g_io_channel_read_line(channel, &string, &size, NULL, NULL);
    if (status == G_IO_STATUS_NORMAL) {
        if (sscanf(string, "DBUSNAME: %s\n", test_dbusname) == 1)
            cli_debug("Parent found child's dbusname as: %s", test_dbusname);
        g_free(string);
    }

    return TRUE;
}

static void run_dbus_name_vanish_tests(const char *progname)
{
    gint out = -1;
    gboolean ret;
    GPid pid;
    GIOChannel *out_ch;
    gchar *argv[] = {(gchar *)progname, "dbusnametest", NULL};

    main_loop = g_main_loop_new(NULL, FALSE);

    /* Hold active state request as long we're running tests */
    requestSysState("main-req", POWERD_SYS_STATE_ACTIVE, &main_cookie);

    ret = g_spawn_async_with_pipes(NULL, argv, NULL,
                                   G_SPAWN_SEARCH_PATH | G_SPAWN_DO_NOT_REAP_CHILD,
                                   NULL, NULL, &pid, NULL, &out, NULL, NULL);

    if (!ret) {
        cli_warn("Failed to spawn test app");
        return;
    }

    g_child_watch_add(pid, (GChildWatchFunc)cb_child_watch, NULL);

    out_ch = g_io_channel_unix_new(out);
    g_io_add_watch(out_ch, G_IO_IN | G_IO_HUP, (GIOFunc)cb_out_watch, NULL);

    g_main_loop_run(main_loop);
    g_main_loop_unref(main_loop);
}

static void run_sys_tests()
{
    int i;
    powerd_cookie_t cookie, cookies[TEST_NUM_SYS_REQUESTS];
    GArray *requests = NULL;

    silence_errors(TRUE);

    /* Hold active state request as long we're running tests */
    requestSysState("main-req", POWERD_SYS_STATE_ACTIVE, &main_cookie);

    for (i = 0; i < TEST_NUM_SYS_REQUESTS; i++) {
        char name[16];
        snprintf(name, sizeof(name), "test-cookie-%d", i);
        do_test(requestSysState(name, POWERD_SYS_STATE_ACTIVE, &cookies[i]) == TRUE);
    }

    /* Make sure we have at least NUM_REQUESTS + 1 */
    requests = listSysRequests();
    do_test(requests->len >= TEST_NUM_SYS_REQUESTS + 1);
    do_test(check_for_dbus_name(powerd_cli_bus_name, TEST_NUM_SYS_REQUESTS + 1,
                                requests, TRUE));
    g_array_free(requests, TRUE);

    for (i = 0; i < TEST_NUM_SYS_REQUESTS; i++)
        do_test(clearSysState(cookies[i]) == TRUE);

    /* We should still have at least 1 request here */
    requests = listSysRequests();
    do_test(requests->len >= 1);
    do_test(check_for_dbus_name(powerd_cli_bus_name, 1, requests, TRUE));
    g_array_free(requests, TRUE);

    /*
     * Now try and freeing the same cookies again. They should all be
     * invalid and thus the request fails.
     */
    for (i = 0; i < TEST_NUM_SYS_REQUESTS; i++)
        do_test(clearSysState(cookies[i]) == FALSE);

    /* Test releasing an invalid cookie format */
    do_test(clearSysState("bad cookie") == FALSE);

    /* We should still have at least 1 request here */
    requests = listSysRequests();
    do_test(requests->len >= 1);
    do_test(check_for_dbus_name(powerd_cli_bus_name, 1, requests, TRUE));
    g_array_free(requests, TRUE);

    /*cannot request the suspend state, this will fail */
    do_test(requestSysState("test-cookie", POWERD_SYS_STATE_SUSPEND, &cookie) == FALSE);

    /* invalid values */
    do_test(requestSysState("test-cookie", -1, &cookie) == FALSE);
    do_test(requestSysState("test-cookie", POWERD_NUM_POWER_STATES, &cookie) == FALSE);
    do_test(requestSysState("test-cookie", POWERD_NUM_POWER_STATES+1, &cookie) == FALSE);

    /* cleanup */
    do_test(clearSysState(main_cookie) == TRUE);
    do_test(check_for_dbus_name(powerd_cli_bus_name, 0, requests, TRUE));

    silence_errors(FALSE);
}

static void run_display_tests()
{
    int i;
    int state = 0;
    gboolean bright = FALSE;
    powerd_cookie_t cookie, cookies[TEST_NUM_DISP_REQUESTS];
    GArray *requests;
    struct PublicDispRequest pdr = {0,};

    silence_errors(TRUE);

    /* Hold active state request as long we're running tests */
    requestSysState("main-req", POWERD_SYS_STATE_ACTIVE, &main_cookie);

    for (i = 0; i < TEST_NUM_DISP_REQUESTS; i++) {
        char name[16];

        snprintf(name, sizeof(name), "disp-test-%d", i);

        pdr.state = state++;
        pdr.flags = 0;
        if (bright)
            pdr.flags |= POWERD_DISPLAY_FLAG_BRIGHT;

        do_test(requestDisplayState(pdr, name, &cookies[i]) == TRUE);

        if (state >= POWERD_NUM_DISPLAY_STATES)
            state = 0;

        bright = !bright;
    }

    /* Make sure we have at least NUM_REQUESTS */
    requests = listDisplayRequests();
    do_test(requests->len >= TEST_NUM_DISP_REQUESTS);

    /* We should see our PID NUM_REQUESTS times */
    do_test(check_for_dbus_name(powerd_cli_bus_name, TEST_NUM_DISP_REQUESTS, requests, FALSE));
    g_array_free(requests, TRUE);

    for (i = 0; i < TEST_NUM_DISP_REQUESTS; i++)
        do_test(clearDisplayState(cookies[i]) == TRUE);

    /* We aren't holding any more requests with this PID */
    requests = listDisplayRequests();
    do_test(check_for_dbus_name(powerd_cli_bus_name, 0, requests, FALSE));
    g_array_free(requests, TRUE);

    /*
     * Now try and freeing the same cookies again. They should all be
     * invalid and thus the request fails.
     */
    for (i = 0; i < TEST_NUM_DISP_REQUESTS; i++)
        do_test(clearDisplayState(cookies[i]) == FALSE);

    /* Test releasing an invalid cookie format */
    do_test(clearDisplayState("bad cookie") == FALSE);

    /* invalid values */
    pdr.state = -1;
    do_test(requestDisplayState(pdr, "disp-test", &cookie) == FALSE);

    pdr.state = POWERD_NUM_DISPLAY_STATES;
    do_test(requestDisplayState(pdr, "disp-test", &cookie) == FALSE);

    pdr.state = POWERD_DISPLAY_STATE_ON;
    pdr.flags = 0xdeadbeef;
    do_test(requestDisplayState(pdr, "disp-test", &cookie) == FALSE);

    /* updateDisplayState tests */
    pdr.state = POWERD_DISPLAY_STATE_ON;
    pdr.flags = 0;
    do_test(requestDisplayState(pdr, "disp-test", &cookie) == TRUE);

    /* update with same state should be okay */
    do_test(updateDisplayState(pdr, cookie) == TRUE);

    pdr.state = POWERD_DISPLAY_STATE_DONT_CARE;
    do_test(updateDisplayState(pdr, cookie) == TRUE);

    pdr.flags = POWERD_DISPLAY_FLAG_DISABLE_AUTOBRIGHTNESS;
    do_test(updateDisplayState(pdr, cookie) == TRUE);

    pdr.state = -1;
    do_test(updateDisplayState(pdr, cookie) == FALSE);

    do_test(clearDisplayState(cookie) == TRUE);

    /* update with same cookie after clear should fail */
    do_test(updateDisplayState(pdr, cookie) == FALSE);

    /* update with garbage cookie should fail */
    do_test(updateDisplayState(pdr, "bad cookie") == FALSE);

    /* cleanup */
    do_test(clearSysState(main_cookie) == TRUE);
    do_test(check_for_dbus_name(powerd_cli_bus_name, 0, requests, TRUE));

    silence_errors(FALSE);
}

static int run_test(int argc, char *argv[])
{
    GDBusConnection *powerd_cli_bus;

    if (argc != 1) {
        /* XXX: Usage */
        return -EINVAL;
    }

    powerd_cli_bus = g_bus_get_sync(G_BUS_TYPE_SYSTEM, NULL, NULL);
    if (!powerd_cli_bus) {
        fprintf(stderr, "Error: Could not connect to message bus\n");
        return -EIO;
    }

    powerd_cli_bus_name = g_dbus_connection_get_unique_name(powerd_cli_bus);
    if (!powerd_cli_bus_name) {
        fprintf(stderr, "Error: Could not get unique bus name\n");
        return -EIO;
    }

    run_dbus_name_vanish_tests(cli_get_progname());
    run_sys_tests();
    run_display_tests();

    g_object_unref(powerd_cli_bus);
    powerd_cli_bus_name = NULL;

    return 0;
}

const struct cli_cmd test_cmd = {
    .name = "test",
    .help_text = "Run powerd tests.",
    .run = run_test,
};

static int run_dbusnametest(int argc, char *argv[])
{
    GDBusConnection *bus = NULL;
    const char *bus_name = NULL;
    powerd_cookie_t cookie;
    struct PublicDispRequest pdr;

    if (argc != 1) {
        /* XXX: Usage */
        return -EINVAL;
    }

    bus = g_bus_get_sync(G_BUS_TYPE_SYSTEM, NULL, NULL);
    bus_name = g_dbus_connection_get_unique_name(bus);
    cli_debug("Child dbus name is %s", bus_name);

    /* This printf is read by the parent */
    printf("DBUSNAME: %s\n", bus_name);

    /* Grab some requests so we can see that they get cleared later */
    requestSysState("dbusnametest1", POWERD_SYS_STATE_ACTIVE, &cookie);
    requestSysState("dbusnametest2", POWERD_SYS_STATE_ACTIVE, &cookie);
    requestSysState("dbusnametest3", POWERD_SYS_STATE_ACTIVE, &cookie);

    pdr.state = POWERD_DISPLAY_STATE_DONT_CARE;
    pdr.flags = 0;
    requestDisplayState(pdr, "dbusnametest1", &cookie);
    requestDisplayState(pdr, "dbusnametest2", &cookie);
    requestDisplayState(pdr, "dbusnametest3", &cookie);

    g_object_unref(bus);

    /* Exit here without cleanup */
    return 0;
}

const struct cli_cmd dbusnametest_cmd = {
    .name = "test",
    .help_text = NULL, /* Hidden, internal command */
    .run = run_dbusnametest,
};
