/*
 * Copyright 2013 Canonical Ltd.
 *
 * This file is part of powerd.
 *
 * powerd is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3.
 *
 * powerd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <assert.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <limits.h>
#include <unistd.h>
#include <inttypes.h>
#include <sys/types.h>
#include <glib-object.h>
#include <gio/gio.h>
#include <glib.h>
#include <uuid/uuid.h>
#include <powerd.h>

#include "powerd-cli.h"

#define ARRAY_SIZE(a) (sizeof(a) / sizeof((a)[0]))

const char *cli_progname;
GDBusProxy *powerd_proxy = NULL;

GDBusProxy *cli_get_dbus_proxy(void)
{
    return powerd_proxy;
}

const char *cli_get_progname(void)
{
    return cli_progname;
}

gboolean
requestSysState(const char *name, int state,
        powerd_cookie_t *cookie)
{
    GVariant *ret = NULL;
    GError *error = NULL;
    const char *cookie_ptr;
    gboolean success = TRUE;

    if (cookie == NULL) {
        cli_warn("NULL cookie passed to %s", __func__);
        return FALSE;
    }

    ret = g_dbus_proxy_call_sync(powerd_proxy,
            "requestSysState",
            g_variant_new("(si)", name, state),
            G_DBUS_CALL_FLAGS_NONE,
            -1,
            NULL,
            &error);
    if (ret == NULL) {
        cli_warn("requestSysState failed: %s", error->message);
        g_error_free(error);
        return FALSE;
    }

    g_variant_get(ret, "(&s)", &cookie_ptr);
    if (strlen(cookie_ptr) != sizeof(powerd_cookie_t) - 1) {
        cli_warn("Returned cookie has incorrect size");
        success = FALSE;
    } else {
        strncpy(*cookie, cookie_ptr, sizeof(powerd_cookie_t));
        cli_debug("Got cookie: %s", *cookie);
    }

    g_variant_unref(ret);
    return success;
}

gboolean
requestDisplayState(struct PublicDispRequest pdr, const char *name,
        powerd_cookie_t *cookie)
{
    GVariant *ret = NULL;
    GError *error = NULL;
    const char *cookie_ptr;
    gboolean success = TRUE;

    if (cookie == NULL) {
        cli_warn("NULL cookie passed to %s", __func__);
        return FALSE;
    }

    ret = g_dbus_proxy_call_sync(powerd_proxy,
            "requestDisplayState",
            g_variant_new("(siu)", name, pdr.state, pdr.flags),
            G_DBUS_CALL_FLAGS_NONE,
            -1,
            NULL,
            &error);
    if (ret == NULL) {
        cli_warn("requestDisplayState failed: %s", error->message);
        g_error_free(error);
        return FALSE;
    }

    g_variant_get(ret, "(&s)", &cookie_ptr);
    if (strlen(cookie_ptr) != sizeof(powerd_cookie_t) - 1) {
        cli_warn("Returned cookie has incorrect size");
        success = FALSE;
    } else {
        strncpy(*cookie, cookie_ptr, sizeof(powerd_cookie_t));
        cli_debug("Got cookie: %s", *cookie);
    }

    g_variant_unref(ret);
    return success;
}

gboolean updateDisplayState(struct PublicDispRequest pdr,
                            powerd_cookie_t cookie)
{
    GVariant *ret = NULL;
    GError *error = NULL;
    gboolean success = TRUE;

    if (cookie == NULL) {
        cli_warn("NULL cookie passed to %s", __func__);
        return FALSE;
    }

    ret = g_dbus_proxy_call_sync(powerd_proxy,
            "updateDisplayState",
            g_variant_new("(siu)", cookie, pdr.state, pdr.flags),
            G_DBUS_CALL_FLAGS_NONE,
            -1,
            NULL,
            &error);
    if (!ret) {
        cli_warn("updateDisplayState failed: %s", error->message);
        g_error_free(error);
        return FALSE;
    }

    g_variant_unref(ret);
    return success;
}

GArray *listSysRequests(void)
{
    GVariant *ret, *item = NULL;
    GVariantIter *iter = NULL;
    GArray *retarray = g_array_new(FALSE, FALSE, sizeof(struct PublicSysRequest));
    GError *error = NULL;
    struct PublicSysRequest psr;

    ret = g_dbus_proxy_call_sync(powerd_proxy,
            "listSysRequests",
            NULL,
            G_DBUS_CALL_FLAGS_NONE,
            -1,
            NULL,
            &error);
    if (ret == NULL) {
        cli_warn("listSysRequests failed: %s", error->message);
        g_error_free(error);
    }
    else {
        g_variant_get(ret, "(a(ssi))", &iter);
        while ((item = g_variant_iter_next_value (iter))) {
            g_variant_get_child (item, 0, "s", &psr.name);
            g_variant_get_child (item, 1, "s", &psr.owner);
            g_variant_get_child (item, 2, "i", &psr.state);
            g_array_append_val(retarray, psr);
            g_variant_unref(item);
        }
        g_variant_unref(ret);
    }
    return retarray;
}

GArray *listDisplayRequests(void)
{
    GVariant *ret, *item = NULL;
    GVariantIter *iter = NULL;
    GArray *retarray = g_array_new(FALSE, FALSE, sizeof(struct PublicDispRequest));
    GError *error = NULL;
    struct PublicDispRequest pdr;

    ret = g_dbus_proxy_call_sync(powerd_proxy,
            "listDisplayRequests",
            NULL,
            G_DBUS_CALL_FLAGS_NONE,
            -1,
            NULL,
            &error);
    if (ret == NULL) {
        cli_warn("listDisplayRequests failed: %s", error->message);
        g_error_free(error);
    }
    else {
        g_variant_get(ret, "(a(ssiu))", &iter);
        while ((item = g_variant_iter_next_value (iter))) {
            g_variant_get_child(item, 0, "s", &pdr.name);
            g_variant_get_child(item, 1, "s", &pdr.owner);
            g_variant_get_child(item, 2, "i", &pdr.state);
            g_variant_get_child(item, 3, "u", &pdr.flags);
            g_array_append_val(retarray, pdr);
            g_variant_unref(item);
        }
        g_variant_unref(ret);
    }
    return retarray;
}

gboolean
clearSysState(powerd_cookie_t cookie)
{
    GVariant *ret = NULL;
    GError *error = NULL;

    ret = g_dbus_proxy_call_sync(powerd_proxy,
            "clearSysState",
            g_variant_new("(s)", cookie),
            G_DBUS_CALL_FLAGS_NONE,
            -1,
            NULL,
            &error);
    if (ret == NULL) {
        cli_warn("clearSysState failed: %s", error->message);
        g_error_free(error);
        return FALSE;
    }
    g_variant_unref(ret);
    return TRUE;
}

gboolean
clearDisplayState(powerd_cookie_t cookie)
{
    GVariant *ret = NULL;
    GError *error = NULL;

    ret = g_dbus_proxy_call_sync(powerd_proxy,
            "clearDisplayState",
            g_variant_new("(s)", cookie),
            G_DBUS_CALL_FLAGS_NONE,
            -1,
            NULL,
            &error);
    if (ret == NULL) {
        cli_warn("clearDisplayState failed: %s", error->message);
        g_error_free(error);
        return FALSE;
    }
    g_variant_unref(ret);
    return TRUE;
}

static void
usage(const char *progname)
{
    printf("\n\n%s Options:\n",progname);
    printf("active - request the active state, prevent the system from\n"\
           "\tsuspending. This does not modify the screen state.\n");
    printf("active-nc - same as active, but relies on powerd to cleanup\n"\
            "\tthe request. This is only for testing and will immediately"\
            "exit, causing the request to be dropped.\n");
    printf("autobrightness <enable|disable> - enable or disable autobrightness\n");
    printf("brightness <brightness> - set user screen brightness\n");
    printf("brightness-params - get brightness parameters (min, max, etc.)\n");
    printf("clear-sys <cookie> - clear a System state request given a cookie\n");
    printf("clear-disp <cookie> - clear a Display state request given a cookie\n");
    printf("client-test - test powerd registration / ack API\n");
    printf("display <on|dc> [bright] [proximity] [disableab]\n"\
           "\tMake a display state request with the input parameters.\n"\
           "\tThe first argument represents the state of the display:\n"\
           "\tOn, Off, or Don't Care (dc).\n"\
           "\tThe final optional arguments respectively:\n"\
           "\t * make the screen bright [bright]\n"\
           "\t * enable the proximity sensor [proximity]\n"\
           "\t * disable autobrightness [disableab]\n");
    printf("help - display this usage information\n");
    printf("list - list outstanding requests\n");
    printf("listen - listen for signals from powerd. This runs a\n"\
           "\tgmainloop and must be manually killed.\n");
    printf("stats - print request statistics\n");
    printf("test - runs tests\n");
}

static void
sigint_quit(int signal)
{
    /* do nothing */
}

const static struct cli_cmd *cli_cmds[] = {
    &active_cmd,
    &active_nc_cmd,
    &autobrightness_cmd,
    &brightness_cmd,
    &brightness_params_cmd,
    &clear_disp_cmd,
    &clear_sys_cmd,
    &client_test_cmd,
    &dbusnametest_cmd,
    &display_cmd,
    &list_cmd,
    &listen_cmd,
    &stats_cmd,
    &test_cmd,
};

#define NUM_CLI_CMDS ARRAY_SIZE(cli_cmds)

int main (int argc, char **argv)
{
    GError *error = NULL;
    uid_t myeuid;
    const struct cli_cmd *cmd = NULL;
    int i, ret = 0;

    if (argc < 2) {
        fprintf(stderr,"Incorrect number of options\n");
        usage(argv[0]);
        return -1;
    }

    cli_progname = argv[0];

    for (i = 0; i < NUM_CLI_CMDS; i++) {
        if (!strcmp(argv[1], cli_cmds[i]->name)) {
            cmd = cli_cmds[i];
            break;
        }
    }

    if (!cmd &&
       (strcmp(argv[1],"help"))) {
        fprintf(stderr,"Invalid option %s\n",argv[1]);
        usage(argv[0]);
        return -1;
    }

    if (!strcmp(argv[1],"help")) {
        usage(argv[0]);
        return 0;
    }

    myeuid = geteuid();
    if (myeuid != 0) {
        fprintf(stderr,"You must be root to run %s\n",argv[0]);
        return -1;
    }

    signal(SIGINT, sigint_quit);

    powerd_proxy = g_dbus_proxy_new_for_bus_sync(G_BUS_TYPE_SYSTEM,
                G_DBUS_PROXY_FLAGS_NONE,
                NULL,
                "com.canonical.powerd",
                "/com/canonical/powerd",
                "com.canonical.powerd",
                NULL,
                &error);

    if (error != NULL) {
        cli_warn("could not connect to powerd: %s", error->message);
        g_error_free(error);
        return -1;
    }

    if (cmd)
        ret = cmd->run(argc - 1, &argv[1]);

    return (ret == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
}
