/*
 * Copyright 2013 Canonical Ltd.
 *
 * This file is part of powerd.
 *
 * powerd is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3.
 *
 * powerd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __POWERD_CLI_H__
#define __POWERD_CLI_H__

#include <glib.h>
#include <glib-object.h>
#include <gio/gio.h>

#include <powerd.h>

#define do_test(test)                                               \
    ({                                                              \
        gboolean result;                                            \
        printf("Test: " #test "\n");                                \
        result = (test);                                            \
        printf("    result: %s\n", result ? "PASSED" : "FAILED");   \
        result;                                                     \
    })

extern gboolean silent_errors;

#define cli_warn(f, a...)   do { if (!silent_errors) g_warning(f, ##a); } while (0)
#define cli_debug(f, a...)  g_debug(f, ##a)

struct PublicSysRequest {
    const char *name;
    const char *owner;
    int state;
};

struct PublicDispRequest {
    const char *name;
    const char *owner;
    enum powerd_display_state state;
    guint32 flags;
};

/*
 * powerd_cli command definition
 *
 * Every command should create one of these structs and add it to
 * the cli_cmds array.
 *
 * @name: Name of the command on the command-line
 * @help_text: Usage text to display to the user. Set to NULL to omit
 *      from help output (i.e. for internal-only commands).
 * @run: Callback to execute the command. The arguments are the same
 *      as the main() function in a C application, with argv[0]
 *      holding the name of the command (useful if one function is to
 *      be overloaded to support multiple commands).
 */
struct cli_cmd {
    const char *name;
    const char *help_text;
    int (*run)(int argc, char *argv[]);
};

/* cli commands */
const struct cli_cmd active_cmd;
const struct cli_cmd active_nc_cmd;
const struct cli_cmd autobrightness_cmd;
const struct cli_cmd brightness_cmd;
const struct cli_cmd brightness_params_cmd;
const struct cli_cmd clear_disp_cmd;
const struct cli_cmd clear_sys_cmd;
const struct cli_cmd client_test_cmd;
const struct cli_cmd dbusnametest_cmd;
const struct cli_cmd display_cmd;
const struct cli_cmd list_cmd;
const struct cli_cmd listen_cmd;
const struct cli_cmd stats_cmd;
const struct cli_cmd test_cmd;

GDBusProxy *cli_get_dbus_proxy(void);
const char *cli_get_progname(void);

gboolean requestSysState(const char *name, int state, powerd_cookie_t *cookie);
gboolean clearSysState(powerd_cookie_t cookie);

gboolean requestDisplayState(struct PublicDispRequest pdr, const char *name,
                             powerd_cookie_t *cookie);
gboolean clearDisplayState(powerd_cookie_t cookie);
gboolean updateDisplayState(struct PublicDispRequest pdr,
                            powerd_cookie_t cookie);

GArray *listSysRequests(void);
GArray *listDisplayRequests(void);

#endif /* __POWERD_CLI_H__ */
